### Copyright 2023 [Dawn Of Eve]

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Type
from torch.optim.optimizer import Optimizer


from .adadelta import Adadelta, AdadeltaConfig
from .adagrad import Adagrad, AdagradConfig
from .adam import Adam, AdamConfig
from .adamax import Adamax, AdamaxConfig
from .base import BaseOptimizer, BaseConfig
from .lion import Lion, LionConfig
from .momentum import Momentum, MomentumConfig
from .rmsprop import RMSProp, RMSPropConfig
from .radam import Radam, RadamConfig
from .sgd import SGD, SGDConfig


__version__ = "0.0.3"

__all__ = ('Adadelta',
           'AdadeltaConfig',
           'Adagrad',
           'AdagradConfig',
           'Adam',
           'AdamConfig',
           'Adamax', 
           'AdamaxConfig',
           'Adam' 
           'BaseOptimizer',
           'BaseConfig',
           'Lion',
           'LionConfig',
           'Momentum',
           'MomentumConfig',
           'RMSProp',
           'RMSPropConfig',
           'Radam',
           'RadamConfig',
           'SGD',
           'SGDConfig')



### Copyright 2023 [Dawn Of Eve]

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Tuple, Any, Optional
from dataclasses import dataclass

import torch
from torch.optim.optimizer import Optimizer

__all__ = ['BaseConfig', 'BaseOptimizer']

@dataclass
class BaseConfig:
  lr : float = 1E-3
  momentum : bool = False
  adaptive : bool = False
  beta_1 : float = 0.0
  beta_2 : float = 0.0
  eps : float = 1E-8
  weight_decay : float = 0.0
  amsgrad : bool = False

  def dict(self):
    return self.__dict__


class BaseOptimizer (Optimizer):

  def __init__  (self, params, config: BaseConfig = BaseConfig()):
    if not config.lr > 0.0:
      raise ValueError(f"Invalid value for lr in config: {config.lr} ", 
                       "Value must be > 0")
    if not 1.0 > config.beta_1 >= 0.0:
      raise ValueError(f"Invalid value for beta_1 in config: {config.beta_1} ", 
                       "Value must be 1 > x >= 0")
    if not 1.0 > config.beta_2 >= 0.0:
      raise ValueError(f"Invalid value for beta_2 in config: {config.beta_2} ", 
                       "Value must be 1 > x >= 0")
    super().__init__(params, config.dict())

    self.config = config

  def init_state(self,
                 state,
                 group,
                 param):
    state['step'] = 0

    if self.config.momentum:
      state['momentum'] = torch.zeros_like(param, memory_format=torch.preserve_format)
    
    if self.config.adaptive:
      state['adaptivity'] = torch.zeros_like(param, memory_format=torch.preserve_format)
      
      if self.config.amsgrad:
        state['amsgrad'] = torch.zeros_like(param, memory_format=torch.preserve_format)

  def momentum(self,
               state, 
               grad):
    step = state['step']
    m = state['momentum']
    beta_1 = self.config.beta_1

    m.mul_(beta_1).add_(grad, alpha= (1 - beta_1))
    m_hat = m.div(1 - beta_1**(step + 1))

    state['momentum'] = m
    return m_hat

  def amsgrad(adaptivity):

    def __adaptivity__(self, state, grad):
      u = adaptivity(self, state, grad)

      if self.config.amsgrad:
        v = state['amsgrad']
        v = torch.max(v, u)
        state['amsgrad'] = v
        return v
      
      return u

    return __adaptivity__

  @amsgrad
  def adaptivity(self, 
                 state, 
                 grad):
    
    step = state['step']
    v = state['adaptivity']
    beta_2 = self.config.beta_2

    v.mul_(beta_2).addcmul_(grad, grad, value = (1 - beta_2))
    v_hat = v.div(1 - beta_2**(step + 1))

    state['adaptivity'] = v
    return torch.sqrt(v_hat + self.config.eps)

  def update(self,
             state: Dict[str, any],
             group: Dict[str, any],
             grad:  torch.Tensor,
             param: torch.Tensor):
    
    lr = group['lr']

    if self.config.momentum:
      m = self.momentum(state, grad)
      upd = m
    else:
      upd = grad
    
    if self.config.adaptive:
      v = self.adaptivity(state, grad)
      param.data.addcdiv_(upd, v, value = -1 * lr)
    else:
      param.data.add_(upd, alpha = -1 * lr)

    if self.config.weight_decay > 0:
      param.data.add_(param.data,
                      alpha = -1 * lr * self.config.weight_decay)
      
    state['step'] += 1

  @torch.no_grad()
  def step(self, closure = None):
    loss = None
    if closure is not None:
      with torch.enable_grad():
        loss = closure()

    for group in self.param_groups:
      for param in group['params']:
        if param.grad is None:
         continue
        grad = param.grad.data
        if grad.is_sparse:
          raise RuntimeError('This Optimizer does not support sparse gradients,'
                                  ' please consider SparseAdam instead')
        state = self.state[param]
        if len(state) == 0:
          self.init_state(state, group, param)

        self.update(state, group, grad, param)
    
    return loss


### Copyright 2023 [Dawn Of Eve]

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, Any, Optional
from dataclasses import dataclass

import torch

from .base import BaseOptimizer
from .base import BaseConfig


__all__ = ['AdagradConfig', 'Adagrad']

@dataclass
class AdagradConfig(BaseConfig):
  lr : float = 1E-3
  adaptive : bool = True
  eps : float = 1E-8

class Adagrad(BaseOptimizer):

  def __init__ (self, params, config : AdagradConfig = AdagradConfig()):
    if not config.adaptive:
      raise ValueError(f"Invalid value for adaptive in config: {config.adaptive} ", 
                       "Value must be True")
    super().__init__(params, config)
    self.config = config

  def adaptivity(self, 
                 state,
                 grad):
    
    v = state['adaptivity']
    v.add_(torch.pow(grad, 2))
    state['adaptivity'] = v

    return torch.sqrt(v + self.config.eps)


### Copyright 2023 [Dawn Of Eve]

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, Any, Optional
from dataclasses import dataclass

import torch

from .base import BaseOptimizer
from .base import BaseConfig


__all__ = ['SGDConfig', 'SGD']

@dataclass
class SGDConfig(BaseConfig):
  lr : float = 1E-3


class SGD(BaseOptimizer):
  
  def __init__(self, params, config: SGDConfig = SGDConfig()):
    super().__init__(params, config)


### Copyright 2023 [Dawn Of Eve]

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, Any, Optional
from dataclasses import dataclass

import torch

from .base import BaseOptimizer
from .base import BaseConfig


__all__ = ['LionConfig', 'Lion']

@dataclass
class LionConfig(BaseConfig):
  lr : float = 1E-4
  momentum : bool = True
  beta_1 : float = 0.9
  beta_2 : float = 0.99
  eps : float = 1E-8
  weight_decay : float = 0.



class Lion(BaseOptimizer):
  
  def __init__(self, params, config: LionConfig = LionConfig()):
    if not config.momentum:
      raise ValueError(f"Invalid value for momentum in config: {config.momentum} ", 
                       "Value must be True")
    if not 1 > config.beta_1 > 0.:
      raise ValueError(f"Invalid value for beta_1 in config: {config.beta_1} ", 
                       "Value must be between 1 and 0")
    if not 1 > config.beta_2 > 0.:
      raise ValueError(f"Invalid value for beta_2 in config: {config.beta_2} ", 
                       "Value must be between 1 and 0")
    super().__init__(params, config)
    self.config = config

  def momentum(self, state, grad):
    m = state['momentum']
    beta_1 = self.config.beta_1
    beta_2 = self.config.beta_2

    u = m.mul(beta_1).add_(grad, alpha=(1-beta_1))
    
    m.mul_(beta_2).add_(grad, alpha=(1-beta_2))

    state['momentum'] = m

    return torch.sign(u)


### Copyright 2023 [Dawn Of Eve]

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, Any, Optional
from dataclasses import dataclass

import torch

from .base import BaseOptimizer
from .base import BaseConfig


__all__ = ['AdadeltaConfig', 'Adadelta']

@dataclass
class AdadeltaConfig(BaseConfig):
  lr : float = 1
  adaptive : bool = True
  rho : float = 0.90
  beta_2 : float = 0.90
  eps : float = 1E-6


class Adadelta(BaseOptimizer):
  def __init__ (self, params, config : AdadeltaConfig = AdadeltaConfig()):
    super().__init__(params, config)

    self.config = config
    if self.config.rho != self.config.beta_2:
      self.config.beta_2 = self.config.rho
  
  def init_state(self, state, group, param):
    state['adaptive_step'] = 0
    state['adaptivity'] = torch.zeros_like(param, memory_format=torch.preserve_format)
    state['acc_delta'] = torch.zeros_like(param, memory_format=torch.preserve_format)

  def update(self, state, group, grad, param):
    eps = self.config.eps
    rho = self.config.rho
    lr = group['lr']
    m = state['acc_delta']

    denom = self.adaptivity(state, grad)

    delta = m.add(eps).sqrt_().div_(denom).mul_(grad)
    
    param.data.add_(delta, alpha = -1 * lr)

    m.mul_(rho).addcmul_(delta, delta, value=(1 - rho))
    state['acc_delta'] = m 


### Copyright 2023 [Dawn Of Eve]

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, Any, Optional
from dataclasses import dataclass

import torch
import math

from .base import BaseOptimizer
from .base import BaseConfig


__all__ = ['RadamConfig', 'Radam']

@dataclass
class RadamConfig(BaseConfig):
  lr : float = 3E-4
  momentum : bool = True
  adaptive : bool = True
  beta_1 : float = 0.9
  beta_2 : float = 0.99
  eps : float = 1E-8
  weight_decay : float = 0.

class Radam(BaseOptimizer):
  def __init__ (self, params, config : RadamConfig = RadamConfig()):
    super().__init__(params, config)
    self.config = config

  def update(self, state, group, grad, param):
    lr = group['lr']
    beta_2 = self.config.beta_2
    step = state['step']
    p_inf = 2/(1-beta_2) - 1

    m = self.momentum(state, grad)
    v = self.adaptivity(state, grad)

    p_t = p_inf - 2 * step * (beta_2 ** (step+1))/(1 - beta_2**(step+1))

    if p_t > 4:
      numrt = ((p_t - 4) * (p_t - 2) * p_inf)
      denomrt = ((p_inf - 4) * (p_inf - 2) * p_t)
      r_t  = math.sqrt(numrt/denomrt)

      param.data.addcdiv_(m, v, value= -1 * r_t * lr)
    else:
      param.data.add_(m, alpha= -lr)

    if self.config.weight_decay > 0:
      param.data.add_(param.data,
                      alpha = -1 * lr * self.config.weight_decay)
    
    state['step'] += 1


### Copyright 2023 [Dawn Of Eve]

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Tuple, Any, Optional
from dataclasses import dataclass

import torch
from torch.optim.optimizer import Optimizer

from .base import BaseOptimizer
from .base import BaseConfig

__all__ = ['AMSGradConfig', 'AMSGrad']

@dataclass
class AMSGradConfig(BaseConfig):
  lr : float = 3E-4
  momentum : bool = True
  adaptive : bool = True
  beta_1 : float = 0.9
  beta_2 : float = 0.999
  eps : float = 1E-8
  weight_decay : float = 0.
  amsgrad : bool = True

class AMSGrad(BaseOptimizer):
  def __init__ (self, params, config : AMSGradConfig = AMSGradConfig()):
    if not config.momentum:
      raise ValueError(f"Invalid value for momentum in config: {config.momentum} ", 
                       "Value must be True")
    if not config.adaptive:
      raise ValueError(f"Invalid value for adaptive in config: {config.adaptive} ", 
                       "Value must be True")
    if not config.amsgrad:
      raise ValueError(f"Invalid value for amsgrad in config: {config.amsgrad} ", 
                       "Value must be True")
    super().__init__(params, config)
    
    self.config = config